from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    RDFS,
)
from rdflib.term import BNode

from fdk_rdf_parser.classes import (
    Reference,
    ReferenceDataCode,
    SkosConcept,
)
from fdk_rdf_parser.rdf_utils import (
    resource_list,
    value_translations,
)


def extract_references(graph: Graph, subject: URIRef) -> Optional[List[Reference]]:
    reference_properties = [
        DCTERMS.hasVersion,
        DCTERMS.isVersionOf,
        DCTERMS.isPartOf,
        DCTERMS.hasPart,
        DCTERMS.references,
        DCTERMS.isReferencedBy,
        DCTERMS.replaces,
        DCTERMS.isReplacedBy,
        DCTERMS.requires,
        DCTERMS.isRequiredBy,
        DCTERMS.relation,
        DCTERMS.source,
    ]

    values = []

    for predicate in reference_properties:
        resources = resource_list(graph, subject, predicate)
        if resources:
            for resource in resources:
                values.append(
                    Reference(
                        referenceType=ReferenceDataCode(uri=predicate.toPython()),
                        source=SkosConcept(
                            uri=(
                                resource.toPython()
                                if not isinstance(resource, BNode)
                                else None
                            ),
                            prefLabel=value_translations(graph, resource, RDFS.label),
                        ),
                    )
                )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

from .skos_concept import SkosConcept


@dataclass
class Requirement:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    dctTitle: Optional[Dict[str, str]] = None
    dctType: Optional[List[SkosConcept]] = None
    description: Optional[Dict[str, str]] = None
    fulfils: Optional[List[str]] = None



from .fdk_rdf_parser import (
    parse_concept,
    parse_concept_as_dict,
    parse_concepts,
    parse_data_service,
    parse_data_services,
    parse_dataservice_as_dict,
    parse_dataset,
    parse_dataset_as_dict,
    parse_datasets,
    parse_event,
    parse_event_as_dict,
    parse_events,
    parse_information_model,
    parse_information_model_as_dict,
    parse_information_models,
    parse_public_services,
    parse_service,
    parse_service_as_dict,
)



from datetime import (
    date,
    datetime,
    timedelta,
)
import numbers
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
)

from isodate import (
    date_isoformat,
    datetime_isoformat,
)
from rdflib import (
    BNode,
    Graph,
    Literal,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    RDF,
)


def is_type(t: URIRef, graph: Graph, topic: URIRef) -> bool:
    for type_uri_ref in resource_list(graph, topic, RDF.type):
        if type_uri_ref == t:
            return True

    return False


def has_value_on_predicate(graph: Graph, subject: URIRef, predicate: URIRef) -> bool:
    return graph.value(subject, predicate) is not None


def has_literal_value_on_predicate(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> bool:
    value = graph.value(subject, predicate)
    return isinstance(value, Literal)


def object_value(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[Any]:
    value = graph.value(subject, predicate)
    return value.toPython() if value and not isinstance(value, BNode) else None


def duration_string_value(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[str]:
    value = graph.value(subject, predicate)
    return (
        str(value)
        if value
        and not isinstance(value, BNode)
        and isinstance(value.toPython(), timedelta)
        else None
    )


def object_number_value(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[Any]:
    value = graph.value(subject, predicate)
    return (
        value.toPython()
        if not isinstance(value, BNode)
        and isinstance(string_to_float(str(value)), numbers.Number)
        else None
    )


def value_list(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[List[Any]]:
    values = []
    for obj in graph.objects(subject, predicate):
        if not isinstance(obj, BNode):
            values.append(obj.toPython())
    values.sort()
    return values if len(values) > 0 else None


def value_set(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[Set[Any]]:
    values = set()
    for obj in graph.objects(subject, predicate):
        if not isinstance(obj, BNode):
            values.add(obj.toPython())
    return values if len(values) > 0 else None


def value_translations(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[Dict[str, str]]:
    values = {}
    for obj in graph.objects(subject, predicate):
        if obj.language:
            values[obj.language] = obj.toPython()
        else:
            values["nb"] = obj.toPython()
    return values if len(values) > 0 else None


def resource_list(graph: Graph, subject: URIRef, predicate: URIRef) -> List[Any]:
    values = []
    for obj in graph.objects(subject, predicate):
        values.append(obj)
    values.sort()
    return values


def uri_or_identifier_list(graph: Graph, subjects: List[URIRef]) -> Optional[List[str]]:
    values = []
    for subject_ref in subjects:
        value = uri_or_identifier(graph, subject_ref)
        if value:
            values.append(value)
    values.sort()
    return values if len(values) > 0 else None


def uri_or_identifier(graph: Graph, subject: URIRef) -> Optional[str]:
    return (
        subject.toPython()
        if isinstance(subject, URIRef)
        else object_value(graph, subject, DCTERMS.identifier)
    )


def date_value(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[str]:
    value = graph.value(subject, predicate)
    if value:
        date_object = value.toPython()
        if isinstance(date_object, datetime):
            return datetime_isoformat(date_object)
        elif isinstance(date_object, date):
            return date_isoformat(date_object)
        else:
            return None
    else:
        return None


def string_to_float(str_value: str) -> Optional[float]:
    dot_split = str_value.split(".")
    comma_split = str_value.split(",")
    if len(dot_split) == 2 and "," not in dot_split[1]:
        to_convert = str_value.replace(",", "")
    elif len(comma_split) == 2 and "." not in comma_split[1]:
        to_convert = str_value.replace(".", "").replace(",", ".")
    else:
        to_convert = str_value

    try:
        return float(to_convert)
    except ValueError:
        return None


linguistic_system_keywords: Dict[str, str] = {
    "http://publications.europa.eu/resource/authority/language/NOR": "no",
    "http://publications.europa.eu/resource/authority/language/NOB": "nb",
    "http://publications.europa.eu/resource/authority/language/NNO": "nn",
    "http://publications.europa.eu/resource/authority/language/ENG": "en",
}



import re
from typing import (
    Dict,
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import PartialDcatResource
from fdk_rdf_parser.rdf_utils import (
    date_value,
    dcat_uri,
    object_value,
    value_list,
    value_set,
    value_translations,
)
from .contactpoint import extract_contact_points
from .publisher import extract_publisher
from .reference_data_code import (
    extract_reference_access_rights,
    extract_reference_language_list,
)
from .theme import (
    map_data_themes,
    map_eurovoc_themes,
    map_los_themes,
    split_theme_refs,
)


def parse_dcat_resource(graph: Graph, subject: URIRef) -> PartialDcatResource:
    formatted_description = value_translations(graph, subject, DCTERMS.description)
    theme_refs = split_theme_refs(graph, subject, dcat_uri("theme"))
    return PartialDcatResource(
        identifier=value_set(graph, subject, DCTERMS.identifier),
        publisher=extract_publisher(graph, subject),
        title=value_translations(graph, subject, DCTERMS.title),
        description=(
            description_html_cleaner(formatted_description)
            if formatted_description
            else None
        ),
        descriptionFormatted=formatted_description,
        uri=subject.toPython(),
        accessRights=extract_reference_access_rights(graph, subject),
        themeUris=value_list(graph, subject, dcat_uri("theme")),
        theme=map_data_themes(graph, theme_refs["data-themes"]),
        losTheme=map_los_themes(graph, theme_refs["los"]),
        eurovocThemes=map_eurovoc_themes(graph, theme_refs["eurovocs"]),
        keyword=extract_key_words(graph, subject),
        contactPoint=extract_contact_points(graph, subject),
        dctType=object_value(graph, subject, DCTERMS.type),
        issued=date_value(graph, subject, DCTERMS.issued),
        modified=date_value(graph, subject, DCTERMS.modified),
        landingPage=value_set(graph, subject, dcat_uri("landingPage")),
        language=extract_reference_language_list(graph, subject, DCTERMS.language),
    )


def extract_key_words(graph: Graph, subject: URIRef) -> Optional[List[Dict[str, str]]]:
    values = []
    for keyword in graph.objects(subject, dcat_uri("keyword")):
        translation = {}
        if keyword.language:
            translation[keyword.language] = keyword.toPython()
        else:
            translation["nb"] = keyword.toPython()
        values.append(translation)
    return values if len(values) > 0 else None


def description_html_cleaner(formatted: Dict[str, str]) -> Optional[Dict[str, str]]:
    cleaner_regex = re.compile("<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});")
    description = {}
    for language in formatted:
        description[language] = re.sub(cleaner_regex, "", formatted[language])

    return description if len(description) > 0 else None



from rdflib.exceptions import ParserError as RdflibParserError


class ParserError(RdflibParserError):
    def __init__(self: "ParserError", msg: str = "Parser error") -> None:
        RdflibParserError.__init__(self, msg)
        self.msg: str = msg


class MultipleResourcesError(ValueError):
    pass


class MissingResourceError(ValueError):
    pass



from typing import (
    TypeVar,
    Union,
)

from fdk_rdf_parser.classes.concept import Concept
from fdk_rdf_parser.classes.cpsvno_service import Service
from fdk_rdf_parser.classes.dataservice import DataService
from fdk_rdf_parser.classes.dataset import Dataset
from fdk_rdf_parser.classes.event import Event
from fdk_rdf_parser.classes.info_model import InformationModel

ResourceType = TypeVar(
    "ResourceType",
    bound=Union[Dataset, DataService, Concept, InformationModel, Event, Service],
)



from typing import (
    List,
    Optional,
)

from rdflib import (
    BNode,
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DC,
    DCTERMS,
    SKOS,
)

from fdk_rdf_parser.classes import ReferenceDataCode
from fdk_rdf_parser.rdf_utils import (
    adms_uri,
    at_uri,
    cv_uri,
    object_value,
    resource_list,
    value_translations,
)


def filter_reference_data_code_list(
    code_list: List[ReferenceDataCode],
) -> Optional[List[ReferenceDataCode]]:
    non_empty_list = []
    for code in code_list:
        if code.uri is not None or code.code is not None or code.prefLabel is not None:
            non_empty_list.append(code)
    return non_empty_list if len(non_empty_list) > 0 else None


def filter_reference_data_code(
    code: ReferenceDataCode,
) -> Optional[ReferenceDataCode]:
    if code.uri is not None or code.code is not None or code.prefLabel is not None:
        return code
    else:
        return None


def parse_reference_code(
    graph: Graph, code_ref: URIRef, code_predicate: URIRef, label_predicate: URIRef
) -> ReferenceDataCode:
    return ReferenceDataCode(
        uri=code_ref.toPython() if not isinstance(code_ref, BNode) else None,
        code=object_value(graph, code_ref, code_predicate),
        prefLabel=value_translations(graph, code_ref, label_predicate),
    )


def extract_reference_language_list(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[ReferenceDataCode]]:
    lang_list = []
    ref_list = resource_list(graph, subject, predicate)
    for lang_ref in ref_list:
        lang_list.append(
            parse_reference_code(
                graph, lang_ref, at_uri("authority-code"), SKOS.prefLabel
            )
        )
    return filter_reference_data_code_list(lang_list)


def extract_reference_language_code(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[ReferenceDataCode]:
    lang_ref = graph.value(subject, predicate)
    if lang_ref:
        return filter_reference_data_code(
            parse_reference_code(graph, lang_ref, DC.identifier, SKOS.prefLabel)
        )
    else:
        return None


def extract_reference_access_rights(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    rights_ref = graph.value(subject, DCTERMS.accessRights)
    if rights_ref:
        rights = parse_reference_code(graph, rights_ref, DC.identifier, SKOS.prefLabel)
        return filter_reference_data_code(rights)
    else:
        return None


def extract_reference_frequency(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    freq_ref = graph.value(subject, DCTERMS.accrualPeriodicity)
    if freq_ref:
        frequency = parse_reference_code(graph, freq_ref, DC.identifier, SKOS.prefLabel)
        return filter_reference_data_code(frequency)
    else:
        return None


def extract_reference_provenance(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    prov_ref = graph.value(subject, DCTERMS.provenance)
    if prov_ref:
        provenance = parse_reference_code(
            graph, prov_ref, at_uri("authority-code"), SKOS.prefLabel
        )
        return filter_reference_data_code(provenance)
    else:
        return None


def extract_reference_adms_status(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    status_ref = graph.value(subject, adms_uri("status"))
    if status_ref:
        status = parse_reference_code(graph, status_ref, SKOS.notation, SKOS.prefLabel)
        return filter_reference_data_code(status)
    else:
        return None


def extract_reference_evidence_types(
    graph: Graph, subject: URIRef
) -> Optional[List[ReferenceDataCode]]:
    evidence_list = []
    ref_list = resource_list(graph, subject, DCTERMS.type)
    for evidence_ref in ref_list:
        evidence = parse_reference_code(
            graph, evidence_ref, DCTERMS.identifier, SKOS.prefLabel
        )
        evidence.code = evidence.code.split("#")[-1] if evidence.code else None
        evidence_list.append(evidence)
    return filter_reference_data_code_list(evidence_list)


def extract_reference_role_types(
    graph: Graph, subject: URIRef
) -> Optional[List[ReferenceDataCode]]:
    role_list = []
    ref_list = resource_list(graph, subject, cv_uri("role"))
    for role_ref in ref_list:
        role = parse_reference_code(graph, role_ref, DCTERMS.identifier, SKOS.prefLabel)
        role.code = role.code.split("#")[-1] if role.code else None
        role_list.append(role)
    return filter_reference_data_code_list(role_list)


def extract_reference_channel_type(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    channel_ref = graph.value(subject, DCTERMS.type)
    if channel_ref:
        channel = parse_reference_code(
            graph, channel_ref, DCTERMS.identifier, SKOS.prefLabel
        )
        channel.code = channel.code.split("#")[-1] if channel.code else None
        return filter_reference_data_code(channel)
    else:
        return None


def extract_reference_publisher_type(
    graph: Graph, subject: URIRef
) -> Optional[ReferenceDataCode]:
    publisher_type_ref = graph.value(subject, DCTERMS.type)
    if publisher_type_ref:
        return filter_reference_data_code(
            parse_reference_code(
                graph, publisher_type_ref, SKOS.notation, SKOS.prefLabel
            )
        )
    else:
        return None


def extract_reference_main_activities(
    graph: Graph, subject: URIRef
) -> Optional[List[ReferenceDataCode]]:
    activities = []
    ref_list = resource_list(graph, subject, DCTERMS.type)
    for activity_ref in ref_list:
        activity = parse_reference_code(
            graph, activity_ref, DC.identifier, SKOS.prefLabel
        )
        activities.append(activity)
    return filter_reference_data_code_list(activities)
